/*
 * Video Analyzer : Simple Media Prober based on the FFmpeg libraries
 *
 * Shared linking
 * gcc -o video_analyzer video_analyzer.c -I/usr/local/include -L/usr/local/lib -lavformat -lavcodec -lavutil
 *
 * Static linking
 * gcc -o video_analyzer video_analyzer.c -I/usr/local/include -L/usr/local/lib -static -lavformat -lavcodec -lavutil -lrtmp -lmatroska -ldirac_decoder -ldirac_encoder -lfaac -lfaad -lmp3lame -ltheora -ltheoradec -ltheoraenc -lvorbis -lvorbisenc -lx264 -lxvidcore -logg -lpthread -lssl -lcrypto -lstdc++ -lm -lz -ldl
 *
 */

#include "libavformat/avformat.h"
#include "libavcodec/avcodec.h"
#include "libavcodec/opt.h"
#include "libavutil/pixdesc.h"

static int show_value_unit              = 0;
static int use_value_prefix             = 0;
static int use_byte_value_binary_prefix = 0;
static int use_value_sexagesimal_format = 0;

static const char *input_filename = NULL;
static AVInputFormat *iformat = NULL;

static const char *binary_unit_prefixes [] = { "", "Ki", "Mi", "Gi", "Ti", "Pi" };
static const char *decimal_unit_prefixes[] = { "", "K" , "M" , "G" , "T" , "P"  };

static const char *unit_second_str          = "s"    ;
static const char *unit_hertz_str           = "Hz"   ;
static const char *unit_byte_str            = "byte" ;
static const char *unit_bit_per_second_str  = "bit/s";

void print_error(const char *filename, int err)
{
    char errbuf[128];
    const char *errbuf_ptr = errbuf;

    if (av_strerror(err, errbuf, sizeof(errbuf)) < 0)
        errbuf_ptr = strerror(AVUNERROR(err));
    fprintf(stderr, "%s: %s\n", filename, errbuf_ptr);
}

static char *value_string(char *buf, int buf_size, double val, const char *unit)
{
    if (unit == unit_second_str && use_value_sexagesimal_format) {
        double secs;
        int hours, mins;
        secs  = val;
        mins  = (int)secs / 60;
        secs  = secs - mins * 60;
        hours = mins / 60;
        mins %= 60;
        snprintf(buf, buf_size, "%d:%02d:%09.6f", hours, mins, secs);
    } else if (use_value_prefix) {
        const char *prefix_string;
        int index;

        if (unit == unit_byte_str && use_byte_value_binary_prefix) {
            index = (int) (log(val)/log(2)) / 10;
            index = av_clip(index, 0, FF_ARRAY_ELEMS(binary_unit_prefixes) -1);
            val /= pow(2, index*10);
            prefix_string = binary_unit_prefixes[index];
        } else {
            index = (int) (log10(val)) / 3;
            index = av_clip(index, 0, FF_ARRAY_ELEMS(decimal_unit_prefixes) -1);
            val /= pow(10, index*3);
            prefix_string = decimal_unit_prefixes[index];
        }

        snprintf(buf, buf_size, "%.3f %s%s", val, prefix_string, show_value_unit ? unit : "");
    } else {
        snprintf(buf, buf_size, "%f %s", val, show_value_unit ? unit : "");
    }

    return buf;
}

static char *time_value_string(char *buf, int buf_size, int64_t val, const AVRational *time_base)
{
    if (val == AV_NOPTS_VALUE) {
        snprintf(buf, buf_size, "N/A");
    } else {
        value_string(buf, buf_size, val * av_q2d(*time_base), unit_second_str);
    }

    return buf;
}

static const char *media_type_string(enum AVMediaType media_type)
{
    switch (media_type) {
    case AVMEDIA_TYPE_VIDEO:      return "video";
    case AVMEDIA_TYPE_AUDIO:      return "audio";
    case AVMEDIA_TYPE_DATA:       return "data";
    case AVMEDIA_TYPE_SUBTITLE:   return "subtitle";
    case AVMEDIA_TYPE_ATTACHMENT: return "attachment";
    default:                      return "unknown";
    }
}

static void show_stream(AVFormatContext *fmt_ctx, int stream_idx)
{
    AVStream *stream = fmt_ctx->streams[stream_idx];
    AVCodecContext *dec_ctx;
    AVCodec *dec;
    char val_str[128];
    AVMetadataTag *tag = NULL;
    AVRational display_aspect_ratio;

    printf("{");

    if ((dec_ctx = stream->codec))
    {
        if ((dec = dec_ctx->codec))
        {
            printf("\"codec_name\":\"%s\",",         dec->name);
            printf("\"codec_long_name\":\"%s\",",    dec->long_name);
        }
        else
            printf("\"codec_name\":\"unknown\",");

        printf("\"codec_type\":\"%s\",",         media_type_string(dec_ctx->codec_type));
        printf("\"codec_time_base\":\"%d/%d\",", dec_ctx->time_base.num, dec_ctx->time_base.den);

        /* print AVI/FourCC tag */
        av_get_codec_tag_string(val_str, sizeof(val_str), dec_ctx->codec_tag);
        printf("\"codec_tag_string\":\"%s\",", val_str);
        printf("\"codec_tag\":\"0x%04x\",", dec_ctx->codec_tag);

        switch (dec_ctx->codec_type)
        {
        case AVMEDIA_TYPE_VIDEO:
            printf("\"width\":\"%d\",",                   dec_ctx->width);
            printf("\"height\":\"%d\",",                  dec_ctx->height);
            printf("\"has_b_frames\":\"%d\",",            dec_ctx->has_b_frames);
            if (dec_ctx->sample_aspect_ratio.num)
            {
                printf("\"sample_aspect_ratio\":\"%d:%d\",", dec_ctx->sample_aspect_ratio.num,
                                                      dec_ctx->sample_aspect_ratio.den);
                av_reduce(&display_aspect_ratio.num, &display_aspect_ratio.den,
                          dec_ctx->width  * dec_ctx->sample_aspect_ratio.num,
                          dec_ctx->height * dec_ctx->sample_aspect_ratio.den,
                          1024*1024);
                printf("\"display_aspect_ratio\":\"%d:%d\",", display_aspect_ratio.num,
                                                       display_aspect_ratio.den);
            }
            printf("\"pix_fmt\":\"%s\",",                 dec_ctx->pix_fmt != PIX_FMT_NONE ?
                   av_pix_fmt_descriptors[dec_ctx->pix_fmt].name : "unknown");
            break;

        case AVMEDIA_TYPE_AUDIO:
            printf("\"sample_rate\":\"%s\",",             value_string(val_str, sizeof(val_str),
                                                                dec_ctx->sample_rate,
                                                                unit_hertz_str));
            printf("\"channels\":\"%d\",",                dec_ctx->channels);
            printf("\"bits_per_sample\":\"%d\",",         av_get_bits_per_sample(dec_ctx->codec_id));
            break;
        }
    }
    else
        printf("\"codec_type\":\"unknown\",");

    if (fmt_ctx->iformat->flags & AVFMT_SHOW_IDS)
        printf("\"id\":\"0x%x\",", stream->id);

    printf("\"r_frame_rate\":\"%d/%d\",",         stream->r_frame_rate.num,   stream->r_frame_rate.den);
    printf("\"avg_frame_rate\":\"%d/%d\",",       stream->avg_frame_rate.num, stream->avg_frame_rate.den);
    printf("\"time_base\":\"%d/%d\",",            stream->time_base.num,      stream->time_base.den);
    if (stream->language[0])
        printf("\"language\":\"%s\",",            stream->language);
    printf("\"start_time\":\"%s\",",   time_value_string(val_str, sizeof(val_str), stream->start_time,
                                                  &stream->time_base));
    printf("\"duration\":\"%s\",",     time_value_string(val_str, sizeof(val_str), stream->duration,
                                                  &stream->time_base));
    if (stream->nb_frames)
        printf("\"nb_frames\":\"%lld\",",    stream->nb_frames);

    if (dec_ctx->bit_rate) {
        printf("\"bitrate\":\"%d\"", dec_ctx->bit_rate / 1000);
    } else {
        printf("\"bitrate\":\"N/A\"");
    }

    printf("}");
}

static void show_format(AVFormatContext *fmt_ctx)
{
    AVMetadataTag *tag = NULL;
    char val_str[128];

    printf("\"muxer\":{");

    printf("\"filename\":\"%s\",",         fmt_ctx->filename);
    printf("\"nb_streams\":\"%d\",",       fmt_ctx->nb_streams);
    printf("\"format_name\":\"%s\",",      fmt_ctx->iformat->name);
    printf("\"format_long_name\":\"%s\",", fmt_ctx->iformat->long_name);
    printf("\"start_time\":\"%s\",",       time_value_string(val_str, sizeof(val_str), fmt_ctx->start_time,
                                                      &AV_TIME_BASE_Q));
    printf("\"duration\":\"%s\",",         time_value_string(val_str, sizeof(val_str), fmt_ctx->duration,
                                                      &AV_TIME_BASE_Q));
    printf("\"size\":\"%s\",",             value_string(val_str, sizeof(val_str), fmt_ctx->file_size,
                                                 unit_byte_str));
    printf("\"bit_rate\":\"%s\"",         value_string(val_str, sizeof(val_str), fmt_ctx->bit_rate,
                                                 unit_bit_per_second_str));
    printf("}");
}

static int open_input_file(AVFormatContext **fmt_ctx_ptr, const char *filename)
{
    int err, i;
    AVFormatContext *fmt_ctx;

    fmt_ctx = avformat_alloc_context();

    if ((err = av_open_input_file(&fmt_ctx, filename, iformat, 0, NULL)) < 0) {
        print_error(filename, err);
        return err;
    }

    /* fill the streams in the format context */
    if ((err = av_find_stream_info(fmt_ctx)) < 0) {
        print_error(filename, err);
        return err;
    }

//     dump_format(fmt_ctx, 0, filename, 0);

    /* bind a decoder to each input stream */
    for (i = 0; i < fmt_ctx->nb_streams; i++) {
        AVStream *stream = fmt_ctx->streams[i];
        AVCodec *codec;

        if (!(codec = avcodec_find_decoder(stream->codec->codec_id))) {
            fprintf(stderr, "Unsupported codec (id=%d) for input stream %d\n",
                    stream->codec->codec_id, stream->index);
        } else if (avcodec_open(stream->codec, codec) < 0) {
            fprintf(stderr, "Error while opening codec for input stream %d\n",
                    stream->index);
        }
    }

    *fmt_ctx_ptr = fmt_ctx;
    return 0;
}

static int probe_file(const char *filename)
{
    AVFormatContext *fmt_ctx;
    int ret, i;

    if ((ret = open_input_file(&fmt_ctx, filename)))
        return ret;

    printf("{");
    show_format(fmt_ctx);

    printf(",\"streams\":[");
    for (i = 0; i < fmt_ctx->nb_streams; i++)
    {
        show_stream(fmt_ctx, i);

        if ( i < (fmt_ctx->nb_streams - 1) )
            printf(",");
    }

    printf("]}");
    av_close_input_file(fmt_ctx);
    return 0;
}

static void opt_input_file(const char *arg)
{
    if (input_filename)
    {
        fprintf(stderr, "Argument '%s' provided as input filename, but '%s' was already specified.\n",
                arg, input_filename);
        exit(1);
    }
    if (!strcmp(arg, "-"))
        arg = "pipe:";
    input_filename = arg;
}

static void show_usage(void)
{
    printf("Simple multimedia streams analyzer\n");
    printf("usage: video analyzer [INPUT_FILE]\n");
    printf("\n");
}

int main(int argc, char **argv)
{
    av_register_all();

    if ( argc > 1 )
        input_filename = argv[1];

    if (!input_filename)
    {
        show_usage();
        fprintf(stderr, "You have to specify one input file.\n");
        exit(1);
    }

    return probe_file(input_filename);
}
